@@ -1,4 +1,4 @@
-// Copyright 2020 The Grin Developers
+// Copyright 2021 The Grin Developers
 //
 // Licensed under the Apache License, Version 2.0 (the "License");
 // you may not use this file except in compliance with the License.
@@ -15,12 +15,10 @@
 //! Facade and handler for the rest of the blockchain implementation
 //! and mostly the chain pipeline.
 
-use crate::core::core::hash::{Hash, Hashed, ZERO_HASH};
 use crate::core::core::merkle_proof::MerkleProof;
-use crate::core::core::verifier_cache::VerifierCache;
 use crate::core::core::{
 	Block, BlockHeader, BlockSums, Committed, Inputs, KernelFeatures, Output, OutputIdentifier,
-	SegmentIdentifier, Transaction, TxKernel,
+	Transaction, TxKernel,
 };
 use crate::core::global;
 use crate::core::pow;
@@ -29,21 +27,24 @@ use crate::error::{Error, ErrorKind};
 use crate::pipe;
 use crate::store;
 use crate::txhashset;
-use crate::txhashset::{PMMRHandle, Segmenter, TxHashSet};
+use crate::txhashset::{Desegmenter, PMMRHandle, Segmenter, TxHashSet};
 use crate::types::{
 	BlockStatus, ChainAdapter, CommitPos, NoStatus, Options, Tip, TxHashsetWriteStatus,
 };
 use crate::util::secp::pedersen::{Commitment, RangeProof};
 use crate::util::RwLock;
-use crate::ChainStore;
-use grin_core::ser;
+use crate::{
+	core::core::hash::{Hash, Hashed},
+	store::Batch,
+	txhashset::{ExtensionPair, HeaderExtension},
+};
 use grin_store::Error::NotFoundErr;
+use std::collections::HashMap;
 use std::fs::{self, File};
 use std::path::{Path, PathBuf};
 use std::sync::atomic::{AtomicUsize, Ordering};
 use std::sync::Arc;
 use std::time::{Duration, Instant};
-use std::{collections::HashMap, io::Cursor};
 
 /// Orphan pool size is limited by MAX_ORPHAN_SIZE
 pub const MAX_ORPHAN_SIZE: usize = 200;
@@ -151,11 +152,11 @@ pub struct Chain {
 	orphans: Arc<OrphanBlockPool>,
 	txhashset: Arc<RwLock<txhashset::TxHashSet>>,
 	header_pmmr: Arc<RwLock<txhashset::PMMRHandle<BlockHeader>>>,
-	sync_pmmr: Arc<RwLock<txhashset::PMMRHandle<BlockHeader>>>,
-	verifier_cache: Arc<RwLock<dyn VerifierCache>>,
 	pibd_segmenter: Arc<RwLock<Option<Segmenter>>>,
+	pibd_desegmenter: Arc<RwLock<Option<Desegmenter>>>,
 	// POW verification function
 	pow_verifier: fn(&BlockHeader) -> Result<(), pow::Error>,
+	denylist: Arc<RwLock<Vec<Hash>>>,
 	archive_mode: bool,
 	genesis: BlockHeader,
 }
@@ -169,15 +170,10 @@ impl Chain {
 		adapter: Arc<dyn ChainAdapter + Send + Sync>,
 		genesis: Block,
 		pow_verifier: fn(&BlockHeader) -> Result<(), pow::Error>,
-		verifier_cache: Arc<RwLock<dyn VerifierCache>>,
 		archive_mode: bool,
 	) -> Result<Chain, Error> {
 		let store = Arc::new(store::ChainStore::new(&db_root)?);
 
-		// DB migrations to be run prior to the chain being used.
-		// Migrate full blocks to protocol version v3.
-		Chain::migrate_db_v2_v3(&store)?;
-
 		// open the txhashset, creating a new one if necessary
 		let mut txhashset = txhashset::TxHashSet::open(db_root.clone(), store.clone(), None)?;
 
@@ -187,20 +183,8 @@ impl Chain {
 			ProtocolVersion(1),
 			None,
 		)?;
-		let mut sync_pmmr = PMMRHandle::new(
-			Path::new(&db_root).join("header").join("sync_head"),
-			false,
-			ProtocolVersion(1),
-			None,
-		)?;
 
-		setup_head(
-			&genesis,
-			&store,
-			&mut header_pmmr,
-			&mut sync_pmmr,
-			&mut txhashset,
-		)?;
+		setup_head(&genesis, &store, &mut header_pmmr, &mut txhashset)?;
 
 		// Initialize the output_pos index based on UTXO set
 		// and NRD kernel_pos index based recent kernel history.
@@ -218,41 +202,69 @@ impl Chain {
 			orphans: Arc::new(OrphanBlockPool::new()),
 			txhashset: Arc::new(RwLock::new(txhashset)),
 			header_pmmr: Arc::new(RwLock::new(header_pmmr)),
-			sync_pmmr: Arc::new(RwLock::new(sync_pmmr)),
 			pibd_segmenter: Arc::new(RwLock::new(None)),
+			pibd_desegmenter: Arc::new(RwLock::new(None)),
 			pow_verifier,
-			verifier_cache,
+			denylist: Arc::new(RwLock::new(vec![])),
 			archive_mode,
 			genesis: genesis.header,
 		};
 
-		// If known bad block exists on "current chain" then rewind prior to this.
-		// Suppress any errors here in case we cannot find
-		chain.rewind_bad_block()?;
-
-		let header_head = chain.header_head()?;
-		chain.rebuild_sync_mmr(&header_head)?;
-
 		chain.log_heads()?;
 
-		// Temporarily exercising the initialization process.
-		// Note: This is *really* slow because we are starting from cold.
-		//
-		// This is not required as we will lazily initialize our segmenter as required
-		// once we start receiving PIBD segment requests.
-		// In reality we will do this based on PIBD segment requests.
-		// Initialization (once per 12 hour period) will not be this slow once lmdb and PMMRs
-		// are warmed up.
-		if let Ok(segmenter) = chain.segmenter() {
-			let _ = segmenter.kernel_segment(SegmentIdentifier { height: 9, idx: 0 });
-			let _ = segmenter.bitmap_segment(SegmentIdentifier { height: 9, idx: 0 });
-			let _ = segmenter.output_segment(SegmentIdentifier { height: 11, idx: 0 });
-			let _ = segmenter.rangeproof_segment(SegmentIdentifier { height: 7, idx: 0 });
-		}
-
 		Ok(chain)
 	}
 
+	/// Add provided header hash to our "denylist".
+	/// The header corresponding to any "denied" hash will be rejected
+	/// and the peer subsequently banned.
+	pub fn invalidate_header(&self, hash: Hash) -> Result<(), Error> {
+		self.denylist.write().push(hash);
+		Ok(())
+	}
+
+	/// Reset both head and header_head to the provided header.
+	/// Handles simple rewind and more complex fork scenarios.
+	/// Used by the reset_chain_head owner api endpoint.
+	pub fn reset_chain_head<T: Into<Tip>>(&self, head: T) -> Result<(), Error> {
+		let head = head.into();
+
+		let mut header_pmmr = self.header_pmmr.write();
+		let mut txhashset = self.txhashset.write();
+		let mut batch = self.store.batch()?;
+
+		let header = batch.get_block_header(&head.hash())?;
+
+		// Rewind and reapply blocks to reset the output/rangeproof/kernel MMR.
+		txhashset::extending(
+			&mut header_pmmr,
+			&mut txhashset,
+			&mut batch,
+			|ext, batch| {
+				self.rewind_and_apply_fork(&header, ext, batch)?;
+				batch.save_body_head(&head)?;
+				Ok(())
+			},
+		)?;
+
+		// If the rewind of full blocks was successful then we can rewind the header MMR.
+		// Rewind and reapply headers to reset the header MMR.
+		txhashset::header_extending(&mut header_pmmr, &mut batch, |ext, batch| {
+			self.rewind_and_apply_header_fork(&header, ext, batch)?;
+			batch.save_header_head(&head)?;
+			Ok(())
+		})?;
+
+		batch.commit()?;
+
+		Ok(())
+	}
+
+	/// Are we running with archive_mode enabled?
+	pub fn archive_mode(&self) -> bool {
+		self.archive_mode
+	}
+
 	/// Return our shared header MMR handle.
 	pub fn header_pmmr(&self) -> Arc<RwLock<PMMRHandle<BlockHeader>>> {
 		self.header_pmmr.clone()
@@ -268,97 +280,6 @@ impl Chain {
 		self.store.clone()
 	}
 
-	/// Known bad block that we must rewind prior to if seen on "current chain".
-	fn rewind_bad_block(&self) -> Result<(), Error> {
-		let hash =
-			Hash::from_hex("0002897182d8cf7631e86d56ad546b7cf0893bda811592aa9312ae633ce04813")?;
-
-		if let Ok(header) = self.get_block_header(&hash) {
-			if self.is_on_current_chain(&header).is_ok() {
-				debug!(
-					"rewind_bad_block: found header: {} at {}",
-					header.hash(),
-					header.height
-				);
-
-				let prev_header = self.get_previous_header(&header)?;
-				let new_head = Tip::from_header(&prev_header);
-
-				// Fix the (full) block chain.
-				if let Ok(block) = self.get_block(&hash) {
-					debug!(
-						"rewind_bad_block: found block: {} at {}",
-						block.header.hash(),
-						block.header.height
-					);
-
-					debug!(
-						"rewind_bad_block: rewinding to prev: {} at {}",
-						prev_header.hash(),
-						prev_header.height
-					);
-
-					let mut header_pmmr = self.header_pmmr.write();
-					let mut txhashset = self.txhashset.write();
-					let mut batch = self.store.batch()?;
-
-					let old_head = batch.head()?;
-
-					txhashset::extending(
-						&mut header_pmmr,
-						&mut txhashset,
-						&mut batch,
-						|ext, batch| {
-							pipe::rewind_and_apply_fork(&prev_header, ext, batch)?;
-
-							// Reset chain head.
-							batch.save_body_head(&new_head)?;
-							batch.save_header_head(&new_head)?;
-
-							Ok(())
-						},
-					)?;
-
-					// Cleanup all subsequent bad blocks (back from old head).
-					let mut current = batch.get_block_header(&old_head.hash())?;
-					while current.height > new_head.height {
-						let _ = batch.delete_block(&current.hash());
-						current = batch.get_previous_header(&current)?;
-					}
-
-					batch.commit()?;
-				}
-
-				{
-					let mut header_pmmr = self.header_pmmr.write();
-					let mut batch = self.store.batch()?;
-
-					let old_header_head = batch.header_head()?;
-
-					txhashset::header_extending(&mut header_pmmr, &mut batch, |ext, batch| {
-						pipe::rewind_and_apply_header_fork(&prev_header, ext, batch)?;
-
-						// Reset chain head.
-						batch.save_header_head(&new_head)?;
-
-						Ok(())
-					})?;
-
-					// cleanup all subsequent bad headers (back from old header_head).
-					let mut current = batch.get_block_header(&old_header_head.hash())?;
-					while current.height > new_head.height {
-						let _ = batch.delete_block_header(&current.hash());
-						current = batch.get_previous_header(&current)?;
-					}
-
-					batch.commit()?;
-				}
-			}
-		}
-
-		Ok(())
-	}
-
 	fn log_heads(&self) -> Result<(), Error> {
 		let log_head = |name, head: Tip| {
 			debug!(
@@ -371,7 +292,6 @@ impl Chain {
 		};
 		log_head("head", self.head()?);
 		log_head("header_head", self.header_head()?);
-		log_head("sync_head", self.get_sync_head()?);
 		Ok(())
 	}
 
@@ -386,43 +306,6 @@ impl Chain {
 		res
 	}
 
-	/// We plan to support receiving blocks with CommitOnly inputs.
-	/// We also need to support relaying blocks with FeaturesAndCommit inputs to peers.
-	/// So we need a way to convert blocks from CommitOnly to FeaturesAndCommit.
-	/// Validating the inputs against the utxo_view allows us to look the outputs up.
-	pub fn convert_block_v2(&self, block: Block) -> Result<Block, Error> {
-		debug!(
-			"convert_block_v2: {} at {} ({} -> v2)",
-			block.header.hash(),
-			block.header.height,
-			block.inputs().version_str(),
-		);
-
-		if block.inputs().is_empty() {
-			return Ok(Block {
-				header: block.header,
-				body: block.body.replace_inputs(Inputs::FeaturesAndCommit(vec![])),
-			});
-		}
-
-		let mut header_pmmr = self.header_pmmr.write();
-		let mut txhashset = self.txhashset.write();
-		let inputs: Vec<_> =
-			txhashset::extending_readonly(&mut header_pmmr, &mut txhashset, |ext, batch| {
-				let previous_header = batch.get_previous_header(&block.header)?;
-				pipe::rewind_and_apply_fork(&previous_header, ext, batch)?;
-				ext.extension
-					.utxo_view(ext.header_extension)
-					.validate_inputs(&block.inputs(), batch)
-					.map(|outputs| outputs.into_iter().map(|(out, _)| out).collect())
-			})?;
-		let inputs = inputs.as_slice().into();
-		Ok(Block {
-			header: block.header,
-			body: block.body.replace_inputs(inputs),
-		})
-	}
-
 	fn determine_status(
 		&self,
 		head: Option<Tip>,
@@ -433,7 +316,7 @@ impl Chain {
 		// If head is updated then we are either "next" block or we just experienced a "reorg" to new head.
 		// Otherwise this is a "fork" off the main chain.
 		if let Some(head) = head {
-			if head.prev_block_h == prev_head.last_block_h {
+			if self.is_on_current_chain(prev_head, head).is_ok() {
 				BlockStatus::Next { prev }
 			} else {
 				BlockStatus::Reorg {
@@ -452,7 +335,8 @@ impl Chain {
 	}
 
 	/// Quick check for "known" duplicate block up to and including current chain head.
-	fn is_known(&self, header: &BlockHeader) -> Result<(), Error> {
+	/// Returns an error if this block is "known".
+	pub fn is_known(&self, header: &BlockHeader) -> Result<(), Error> {
 		let head = self.head()?;
 		if head.hash() == header.hash() {
 			return Err(ErrorKind::Unfit("duplicate block".into()).into());
@@ -502,89 +386,45 @@ impl Chain {
 	/// Returns true if it has been added to the longest chain
 	/// or false if it has added to a fork (or orphan?).
 	fn process_block_single(&self, b: Block, opts: Options) -> Result<Option<Tip>, Error> {
-		// Check if we already know about this block.
-		self.is_known(&b.header)?;
-
 		// Process the header first.
 		// If invalid then fail early.
 		// If valid then continue with block processing with header_head committed to db etc.
 		self.process_block_header(&b.header, opts)?;
 
+		// Check if we already know about this full block.
+		self.is_known(&b.header)?;
+
 		// Check if this block is an orphan.
 		// Only do this once we know the header PoW is valid.
 		self.check_orphan(&b, opts)?;
 
-		// We can only reliably convert to "v2" if not an orphan (may spend output from previous block).
-		// We convert from "v3" to "v2" by looking up outputs to be spent.
-		// This conversion also ensures a block received in "v2" has valid input features (prevents malleability).
-		let b = self.convert_block_v2(b)?;
-
-		let (maybe_new_head, prev_head) = {
+		let (head, fork_point, prev_head) = {
 			let mut header_pmmr = self.header_pmmr.write();
 			let mut txhashset = self.txhashset.write();
 			let batch = self.store.batch()?;
 			let prev_head = batch.head()?;
 			let mut ctx = self.new_ctx(opts, batch, &mut header_pmmr, &mut txhashset)?;
 
-			let maybe_new_head = pipe::process_block(&b, &mut ctx);
+			let (head, fork_point) = pipe::process_block(&b, &mut ctx)?;
 
-			// We have flushed txhashset extension changes to disk
-			// but not yet committed the batch.
-			// A node shutdown at this point can be catastrophic...
-			// We prevent this via the stop_lock (see above).
-			if maybe_new_head.is_ok() {
-				ctx.batch.commit()?;
-			}
+			ctx.batch.commit()?;
 
 			// release the lock and let the batch go before post-processing
-			(maybe_new_head, prev_head)
+			(head, fork_point, prev_head)
 		};
 
-		match maybe_new_head {
-			Ok((head, fork_point)) => {
-				let prev = self.get_previous_header(&b.header)?;
-				let status = self.determine_status(
-					head,
-					Tip::from_header(&prev),
-					prev_head,
-					Tip::from_header(&fork_point),
-				);
+		let prev = self.get_previous_header(&b.header)?;
+		let status = self.determine_status(
+			head,
+			Tip::from_header(&prev),
+			prev_head,
+			Tip::from_header(&fork_point),
+		);
 
-				// notifying other parts of the system of the update
-				self.adapter.block_accepted(&b, status, opts);
+		// notifying other parts of the system of the update
+		self.adapter.block_accepted(&b, status, opts);
 
-				Ok(head)
-			}
-			Err(e) => match e.kind() {
-				ErrorKind::InvalidBlockProof(err) => {
-					debug!(
-						"Block {} at {}: block proof error: {:?}",
-						b.hash(),
-						b.header.height,
-						&err
-					);
-					Err(e)
-				}
-				ErrorKind::Unfit(ref msg) => {
-					debug!(
-						"Block {} at {} is unfit at this time: {}",
-						b.hash(),
-						b.header.height,
-						msg
-					);
-					Err(ErrorKind::Unfit(msg.clone()).into())
-				}
-				_ => {
-					info!(
-						"Rejected block {} at {}: {:?}",
-						b.hash(),
-						b.header.height,
-						e
-					);
-					Err(ErrorKind::Other(format!("{:?}", e)).into())
-				}
-			},
-		}
+		Ok(head)
 	}
 
 	/// Process a block header received during "header first" propagation.
@@ -603,28 +443,23 @@ impl Chain {
 	/// Attempt to add new headers to the header chain (or fork).
 	/// This is only ever used during sync and is based on sync_head.
 	/// We update header_head here if our total work increases.
-	pub fn sync_block_headers(&self, headers: &[BlockHeader], opts: Options) -> Result<(), Error> {
-		let mut sync_pmmr = self.sync_pmmr.write();
+	/// Returns the new sync_head (may temporarily diverge from header_head when syncing a long fork).
+	pub fn sync_block_headers(
+		&self,
+		headers: &[BlockHeader],
+		sync_head: Tip,
+		opts: Options,
+	) -> Result<Option<Tip>, Error> {
 		let mut header_pmmr = self.header_pmmr.write();
 		let mut txhashset = self.txhashset.write();
+		let batch = self.store.batch()?;
 
-		// Sync the chunk of block headers, updating sync_head as necessary.
-		{
-			let batch = self.store.batch()?;
-			let mut ctx = self.new_ctx(opts, batch, &mut sync_pmmr, &mut txhashset)?;
-			pipe::sync_block_headers(headers, &mut ctx)?;
-			ctx.batch.commit()?;
-		}
+		// Sync the chunk of block headers, updating header_head if total work increases.
+		let mut ctx = self.new_ctx(opts, batch, &mut header_pmmr, &mut txhashset)?;
+		let sync_head = pipe::process_block_headers(headers, sync_head, &mut ctx)?;
+		ctx.batch.commit()?;
 
-		// Now "process" the last block header, updating header_head to match sync_head.
-		if let Some(header) = headers.last() {
-			let batch = self.store.batch()?;
-			let mut ctx = self.new_ctx(opts, batch, &mut header_pmmr, &mut txhashset)?;
-			pipe::process_block_header(header, &mut ctx)?;
-			ctx.batch.commit()?;
-		}
-
-		Ok(())
+		Ok(sync_head)
 	}
 
 	/// Build a new block processing context.
@@ -635,10 +470,13 @@ impl Chain {
 		header_pmmr: &'a mut txhashset::PMMRHandle<BlockHeader>,
 		txhashset: &'a mut txhashset::TxHashSet,
 	) -> Result<pipe::BlockContext<'a>, Error> {
+		let denylist = self.denylist.read().clone();
 		Ok(pipe::BlockContext {
 			opts,
 			pow_verifier: self.pow_verifier,
-			verifier_cache: self.verifier_cache.clone(),
+			header_allowed: Box::new(move |header| {
+				pipe::validate_header_denylist(header, &denylist)
+			}),
 			header_pmmr,
 			txhashset,
 			batch,
@@ -721,11 +559,11 @@ impl Chain {
 	}
 
 	/// Retrieves an unspent output using its PMMR position
-	pub fn get_unspent_output_at(&self, pos: u64) -> Result<Output, Error> {
+	pub fn get_unspent_output_at(&self, pos0: u64) -> Result<Output, Error> {
 		let header_pmmr = self.header_pmmr.read();
 		let txhashset = self.txhashset.read();
 		txhashset::utxo_view(&header_pmmr, &txhashset, |utxo, _| {
-			utxo.get_unspent_output_at(pos)
+			utxo.get_unspent_output_at(pos0)
 		})
 	}
 
@@ -826,7 +664,7 @@ impl Chain {
 		// latest block header. Rewind the extension to the specified header to
 		// ensure the view is consistent.
 		txhashset::extending_readonly(&mut header_pmmr, &mut txhashset, |ext, batch| {
-			pipe::rewind_and_apply_fork(&header, ext, batch)?;
+			self.rewind_and_apply_fork(&header, ext, batch)?;
 			ext.extension
 				.validate(&self.genesis, fast_validation, &NoStatus, &header)?;
 			Ok(())
@@ -839,7 +677,7 @@ impl Chain {
 		let prev_root =
 			txhashset::header_extending_readonly(&mut header_pmmr, &self.store(), |ext, batch| {
 				let prev_header = batch.get_previous_header(header)?;
-				pipe::rewind_and_apply_header_fork(&prev_header, ext, batch)?;
+				self.rewind_and_apply_header_fork(&prev_header, ext, batch)?;
 				ext.root()
 			})?;
 
@@ -858,7 +696,7 @@ impl Chain {
 		let (prev_root, roots, sizes) =
 			txhashset::extending_readonly(&mut header_pmmr, &mut txhashset, |ext, batch| {
 				let previous_header = batch.get_previous_header(&b.header)?;
-				pipe::rewind_and_apply_fork(&previous_header, ext, batch)?;
+				self.rewind_and_apply_fork(&previous_header, ext, batch)?;
 
 				let extension = &mut ext.extension;
 				let header_extension = &mut ext.header_extension;
@@ -903,7 +741,7 @@ impl Chain {
 		let mut txhashset = self.txhashset.write();
 		let merkle_proof =
 			txhashset::extending_readonly(&mut header_pmmr, &mut txhashset, |ext, batch| {
-				pipe::rewind_and_apply_fork(&header, ext, batch)?;
+				self.rewind_and_apply_fork(&header, ext, batch)?;
 				ext.extension.merkle_proof(out_id, batch)
 			})?;
 
@@ -917,6 +755,34 @@ impl Chain {
 		txhashset.merkle_proof(commit)
 	}
 
+	/// Rewind and apply fork with the chain specific header validation (denylist) rules.
+	/// If we rewind and re-apply a "denied" block then validation will fail.
+	fn rewind_and_apply_fork(
+		&self,
+		header: &BlockHeader,
+		ext: &mut ExtensionPair,
+		batch: &Batch,
+	) -> Result<BlockHeader, Error> {
+		let denylist = self.denylist.read().clone();
+		pipe::rewind_and_apply_fork(header, ext, batch, &|header| {
+			pipe::validate_header_denylist(header, &denylist)
+		})
+	}
+
+	/// Rewind and apply fork with the chain specific header validation (denylist) rules.
+	/// If we rewind and re-apply a "denied" header then validation will fail.
+	fn rewind_and_apply_header_fork(
+		&self,
+		header: &BlockHeader,
+		ext: &mut HeaderExtension,
+		batch: &Batch,
+	) -> Result<(), Error> {
+		let denylist = self.denylist.read().clone();
+		pipe::rewind_and_apply_header_fork(header, ext, batch, &|header| {
+			pipe::validate_header_denylist(header, &denylist)
+		})
+	}
+
 	/// Provides a reading view into the current txhashset state as well as
 	/// the required indexes for a consumer to rewind to a consistent state
 	/// at the provided block hash.
@@ -930,8 +796,9 @@ impl Chain {
 
 		let mut header_pmmr = self.header_pmmr.write();
 		let mut txhashset = self.txhashset.write();
+
 		txhashset::extending_readonly(&mut header_pmmr, &mut txhashset, |ext, batch| {
-			pipe::rewind_and_apply_fork(&header, ext, batch)?;
+			self.rewind_and_apply_fork(&header, ext, batch)?;
 			ext.extension.snapshot(batch)?;
 
 			// prepare the zip
@@ -998,6 +865,43 @@ impl Chain {
 		))
 	}
 
+	/// instantiate desegmenter (in same lazy fashion as segmenter, though this should not be as
+	/// expensive an operation)
+	pub fn desegmenter(&self, archive_header: &BlockHeader) -> Result<Desegmenter, Error> {
+		// Use our cached desegmenter if we have one and the associated header matches.
+		if let Some(d) = self.pibd_desegmenter.read().as_ref() {
+			if d.header() == archive_header {
+				return Ok(d.clone());
+			}
+		}
+		// If no desegmenter or headers don't match init
+		// TODO: (Check whether we can do this.. we *should* be able to modify this as the desegmenter
+		// is in flight and we cross a horizon boundary, but needs more thinking)
+		let desegmenter = self.init_desegmenter(archive_header)?;
+		let mut cache = self.pibd_desegmenter.write();
+		*cache = Some(desegmenter.clone());
+
+		return Ok(desegmenter);
+	}
+
+	/// initialize a desegmenter, which is capable of extending the hashset by appending
+	/// PIBD segments of the three PMMR trees + Bitmap PMMR
+	/// header should be the same header as selected for the txhashset.zip archive
+	fn init_desegmenter(&self, header: &BlockHeader) -> Result<Desegmenter, Error> {
+		debug!(
+			"init_desegmenter: initializing new desegmenter for {} at {}",
+			header.hash(),
+			header.height
+		);
+
+		Ok(Desegmenter::new(
+			self.txhashset(),
+			self.header_pmmr.clone(),
+			header.clone(),
+			self.store.clone(),
+		))
+	}
+
 	/// To support the ability to download the txhashset from multiple peers in parallel,
 	/// the peers must all agree on the exact binary representation of the txhashset.
 	/// This means compacting and rewinding to the exact same header.
@@ -1051,115 +955,31 @@ impl Chain {
 		Ok(())
 	}
 
-	/// Rebuild the sync MMR based on current header_head.
-	/// We rebuild the sync MMR when first entering sync mode so ensure we
-	/// have an MMR we can safely rewind based on the headers received from a peer.
-	pub fn rebuild_sync_mmr(&self, head: &Tip) -> Result<(), Error> {
-		let mut sync_pmmr = self.sync_pmmr.write();
-		let mut batch = self.store.batch()?;
-		let header = batch.get_block_header(&head.hash())?;
-		txhashset::header_extending(&mut sync_pmmr, &mut batch, |ext, batch| {
-			pipe::rewind_and_apply_header_fork(&header, ext, batch)?;
-			Ok(())
-		})?;
-		batch.commit()?;
-		Ok(())
+	/// Finds the "fork point" where header chain diverges from full block chain.
+	/// If we are syncing this will correspond to the last full block where
+	/// the next header is known but we do not yet have the full block.
+	/// i.e. This is the last known full block and all subsequent blocks are missing.
+	pub fn fork_point(&self) -> Result<BlockHeader, Error> {
+		let body_head = self.head()?;
+		let mut current = self.get_block_header(&body_head.hash())?;
+		while !self.is_on_current_chain(&current, body_head).is_ok() {
+			current = self.get_previous_header(&current)?;
+		}
+		Ok(current)
 	}
 
-	/// Check chain status whether a txhashset downloading is needed
-	pub fn check_txhashset_needed(
-		&self,
-		caller: String,
-		hashes: &mut Option<Vec<Hash>>,
-	) -> Result<bool, Error> {
-		let horizon = global::cut_through_horizon() as u64;
-		let body_head = self.head()?;
+	/// Compare fork point to our horizon.
+	/// If beyond the horizon then we cannot sync via recent full blocks
+	/// and we need a state (txhashset) sync.
+	pub fn check_txhashset_needed(&self, fork_point: &BlockHeader) -> Result<bool, Error> {
+		if self.archive_mode() {
+			debug!("check_txhashset_needed: we are running with archive_mode=true, not needed");
+			return Ok(false);
+		}
+
 		let header_head = self.header_head()?;
-		let sync_head = self.get_sync_head()?;
-
-		debug!(
-			"{}: body_head - {}, {}, header_head - {}, {}, sync_head - {}, {}",
-			caller,
-			body_head.last_block_h,
-			body_head.height,
-			header_head.last_block_h,
-			header_head.height,
-			sync_head.last_block_h,
-			sync_head.height,
-		);
-
-		if body_head.total_difficulty >= header_head.total_difficulty {
-			debug!(
-				"{}: no need txhashset. header_head.total_difficulty: {} <= body_head.total_difficulty: {}",
-				caller, header_head.total_difficulty, body_head.total_difficulty,
-			);
-			return Ok(false);
-		}
-
-		let mut oldest_height = 0;
-		let mut oldest_hash = ZERO_HASH;
-
-		// Start with body_head (head of the full block chain)
-		let mut current = self.get_block_header(&body_head.last_block_h);
-		if current.is_err() {
-			error!(
-				"{}: body_head not found in chain db: {} at {}",
-				caller, body_head.last_block_h, body_head.height,
-			);
-			return Ok(false);
-		}
-
-		//
-		// TODO - Investigate finding the "common header" by comparing header_mmr and
-		// sync_mmr (bytes will be identical up to the common header).
-		//
-		// Traverse back through the full block chain from body head until we find a header
-		// that "is on current chain", which is the "fork point" between existing header chain
-		// and full block chain.
-		while let Ok(header) = current {
-			// break out of the while loop when we find a header common
-			// between the header chain and the current body chain
-			if self.is_on_current_chain(&header).is_ok() {
-				oldest_height = header.height;
-				oldest_hash = header.hash();
-				break;
-			}
-
-			current = self.get_previous_header(&header);
-		}
-
-		// Traverse back through the header chain from header_head back to this fork point.
-		// These are the blocks that we need to request in body sync (we have the header but not the full block).
-		if let Some(hs) = hashes {
-			let mut h = self.get_block_header(&header_head.last_block_h);
-			while let Ok(header) = h {
-				if header.height <= oldest_height {
-					break;
-				}
-				hs.push(header.hash());
-				h = self.get_previous_header(&header);
-			}
-		}
-
-		if oldest_height < header_head.height.saturating_sub(horizon) {
-			if oldest_hash != ZERO_HASH {
-				// this is the normal case. for example:
-				// body head height is 1 (and not a fork), oldest_height will be 1
-				// body head height is 0 (a typical fresh node), oldest_height will be 0
-				// body head height is 10,001 (but at a fork with depth 1), oldest_height will be 10,000
-				// body head height is 10,005 (but at a fork with depth 5), oldest_height will be 10,000
-				debug!(
-					"{}: need a state sync for txhashset. oldest block which is not on local chain: {} at {}",
-					caller, oldest_hash, oldest_height,
-				);
-			} else {
-				// this is the abnormal case, when is_on_current_chain() always return Err, and even for genesis block.
-				error!("{}: corrupted storage? state sync is needed", caller);
-			}
-			Ok(true)
-		} else {
-			Ok(false)
-		}
+		let horizon = global::cut_through_horizon() as u64;
+		Ok(fork_point.height < header_head.height.saturating_sub(horizon))
 	}
 
 	/// Clean the temporary sandbox folder
@@ -1211,8 +1031,8 @@ impl Chain {
 		status.on_setup();
 
 		// Initial check whether this txhashset is needed or not
-		let mut hashes: Option<Vec<Hash>> = None;
-		if !self.check_txhashset_needed("txhashset_write".to_owned(), &mut hashes)? {
+		let fork_point = self.fork_point()?;
+		if !self.check_txhashset_needed(&fork_point)? {
 			warn!("txhashset_write: txhashset received but it's not needed! ignored.");
 			return Err(ErrorKind::InvalidTxHashSet("not needed".to_owned()).into());
 		}
@@ -1343,11 +1163,12 @@ impl Chain {
 		header_pmmr: &txhashset::PMMRHandle<BlockHeader>,
 		batch: &store::Batch<'_>,
 	) -> Result<(), Error> {
-		if self.archive_mode {
+		if self.archive_mode() {
 			return Ok(());
 		}
 
-		let horizon = global::cut_through_horizon() as u64;
+		let mut horizon = global::cut_through_horizon() as u64;
+
 		let head = batch.head()?;
 
 		let tail = match batch.tail() {
@@ -1355,7 +1176,16 @@ impl Chain {
 			Err(_) => Tip::from_header(&self.genesis),
 		};
 
-		let cutoff = head.height.saturating_sub(horizon);
+		let mut cutoff = head.height.saturating_sub(horizon);
+
+		// TODO: Check this, compaction selects a different horizon
+		// block from txhashset horizon/PIBD segmenter when using
+		// Automated testing chain
+		let archive_header = self.txhashset_archive_header()?;
+		if archive_header.height < cutoff {
+			cutoff = archive_header.height;
+			horizon = head.height - archive_header.height;
+		}
 
 		debug!(
 			"remove_historical_blocks: head height: {}, tail height: {}, horizon: {}, cutoff: {}",
@@ -1429,13 +1259,14 @@ impl Chain {
 		}
 
 		// If we are not in archival mode remove historical blocks from the db.
-		if !self.archive_mode {
+		if !self.archive_mode() {
 			self.remove_historical_blocks(&header_pmmr, &batch)?;
 		}
 
 		// Make sure our output_pos index is consistent with the UTXO set.
 		txhashset.init_output_pos_index(&header_pmmr, &batch)?;
 
+		// TODO - Why is this part of chain compaction?
 		// Rebuild our NRD kernel_pos index based on recent kernel history.
 		txhashset.init_recent_kernel_pos_index(&header_pmmr, &batch)?;
 
@@ -1475,7 +1306,7 @@ impl Chain {
 		let txhashset = self.txhashset.read();
 		let last_index = match max_pmmr_index {
 			Some(i) => i,
-			None => txhashset.highest_output_insertion_index(),
+			None => txhashset.output_mmr_size(),
 		};
 		let outputs = txhashset.outputs_by_pmmr_index(start_index, max_count, max_pmmr_index);
 		let rangeproofs =
@@ -1504,13 +1335,14 @@ impl Chain {
 			None => self.head_header()?.height,
 		};
 		// Return headers at the given heights
-		let prev_to_start_header =
-			self.get_header_by_height(start_block_height.saturating_sub(1))?;
-		let end_header = self.get_header_by_height(end_block_height)?;
-		Ok((
-			prev_to_start_header.output_mmr_size + 1,
-			end_header.output_mmr_size,
-		))
+		let start_mmr_size = if start_block_height == 0 {
+			0
+		} else {
+			self.get_header_by_height(start_block_height - 1)?
+				.output_mmr_size + 1
+		};
+		let end_mmr_size = self.get_header_by_height(end_block_height)?.output_mmr_size;
+		Ok((start_mmr_size, end_mmr_size))
 	}
 
 	/// Orphans pool size
@@ -1587,58 +1419,6 @@ impl Chain {
 		self.header_pmmr.read().get_header_hash_by_height(height)
 	}
 
-	/// Migrate our local db from v2 to v3.
-	/// "commit only" inputs.
-	fn migrate_db_v2_v3(store: &ChainStore) -> Result<(), Error> {
-		if store.batch()?.is_blocks_v3_migrated()? {
-			// Previously migrated so skipping.
-			debug!("migrate_db_v2_v3: previously migrated, skipping");
-			return Ok(());
-		}
-		let mut total = 0;
-		let mut keys_to_migrate = vec![];
-		for (k, v) in store.batch()?.blocks_raw_iter()? {
-			total += 1;
-
-			// We want to migrate all blocks that cannot be read via v3 protocol version.
-			let block_v3: Result<Block, _> =
-				ser::deserialize(&mut Cursor::new(&v), ProtocolVersion(3));
-			if block_v3.is_err() {
-				let block_v2: Result<Block, _> =
-					ser::deserialize(&mut Cursor::new(&v), ProtocolVersion(2));
-				if block_v2.is_ok() {
-					keys_to_migrate.push(k);
-				}
-			}
-		}
-		debug!(
-			"migrate_db_v2_v3: {} (of {}) blocks to migrate",
-			keys_to_migrate.len(),
-			total,
-		);
-		let mut count = 0;
-		keys_to_migrate
-			.chunks(100)
-			.try_for_each(|keys| {
-				let batch = store.batch()?;
-				for key in keys {
-					batch.migrate_block(&key, ProtocolVersion(2), ProtocolVersion(3))?;
-					count += 1;
-				}
-				batch.commit()?;
-				debug!("migrate_db_v2_v3: successfully migrated {} blocks", count);
-				Ok(())
-			})
-			.and_then(|_| {
-				// Set flag to indicate we have migrated all blocks in the db.
-				// We will skip migration in the future.
-				let batch = store.batch()?;
-				batch.set_blocks_v3_migrated(true)?;
-				batch.commit()?;
-				Ok(())
-			})
-	}
-
 	/// Gets the block header in which a given output appears in the txhashset.
 	pub fn get_header_for_output(&self, commit: Commitment) -> Result<BlockHeader, Error> {
 		let header_pmmr = self.header_pmmr.read();
@@ -1744,31 +1524,34 @@ impl Chain {
 	/// Verifies the given block header is actually on the current chain.
 	/// Checks the header_by_height index to verify the header is where we say
 	/// it is
-	pub fn is_on_current_chain(&self, header: &BlockHeader) -> Result<(), Error> {
-		let chain_header = self.get_header_by_height(header.height)?;
-		if chain_header.hash() == header.hash() {
+	fn is_on_current_chain<T: Into<Tip>>(&self, x: T, head: Tip) -> Result<(), Error> {
+		let x: Tip = x.into();
+		if x.height > head.height {
+			return Err(ErrorKind::Other("not on current chain".to_string()).into());
+		}
+
+		if x.hash() == self.get_header_hash_by_height(x.height)? {
 			Ok(())
 		} else {
 			Err(ErrorKind::Other("not on current chain".to_string()).into())
 		}
 	}
 
-	/// Get the tip of the current "sync" header chain.
-	/// This may be significantly different to current header chain.
-	pub fn get_sync_head(&self) -> Result<Tip, Error> {
-		let hash = self.sync_pmmr.read().head_hash()?;
-		let header = self.store.get_block_header(&hash)?;
-		Ok(Tip::from_header(&header))
-	}
-
 	/// Gets multiple headers at the provided heights.
-	/// Note: Uses the sync pmmr, not the header pmmr.
-	pub fn get_locator_hashes(&self, heights: &[u64]) -> Result<Vec<Hash>, Error> {
-		let pmmr = self.sync_pmmr.read();
-		heights
-			.iter()
-			.map(|h| pmmr.get_header_hash_by_height(*h))
-			.collect()
+	/// Note: This is based on the provided sync_head to support syncing against a fork.
+	pub fn get_locator_hashes(&self, sync_head: Tip, heights: &[u64]) -> Result<Vec<Hash>, Error> {
+		let mut header_pmmr = self.header_pmmr.write();
+		txhashset::header_extending_readonly(&mut header_pmmr, &self.store(), |ext, batch| {
+			let header = batch.get_block_header(&sync_head.hash())?;
+			self.rewind_and_apply_header_fork(&header, ext, batch)?;
+
+			let hashes = heights
+				.iter()
+				.filter_map(|h| ext.get_header_hash_by_height(*h))
+				.collect();
+
+			Ok(hashes)
+		})
 	}
 
 	/// Builds an iterator on blocks starting from the current chain head and
@@ -1792,28 +1575,21 @@ fn setup_head(
 	genesis: &Block,
 	store: &store::ChainStore,
 	header_pmmr: &mut txhashset::PMMRHandle<BlockHeader>,
-	sync_pmmr: &mut txhashset::PMMRHandle<BlockHeader>,
 	txhashset: &mut txhashset::TxHashSet,
 ) -> Result<(), Error> {
 	let mut batch = store.batch()?;
 
-	// Apply the genesis header to header and sync MMRs.
+	// Apply the genesis header to header MMR.
 	{
 		if batch.get_block_header(&genesis.hash()).is_err() {
 			batch.save_block_header(&genesis.header)?;
 		}
 
-		if header_pmmr.last_pos == 0 {
+		if header_pmmr.size == 0 {
 			txhashset::header_extending(header_pmmr, &mut batch, |ext, _| {
 				ext.apply_header(&genesis.header)
 			})?;
 		}
-
-		if sync_pmmr.last_pos == 0 {
-			txhashset::header_extending(sync_pmmr, &mut batch, |ext, _| {
-				ext.apply_header(&genesis.header)
-			})?;
-		}
 	}
 
 	// Make sure our header PMMR is consistent with header_head from db if it exists.
@@ -1844,7 +1620,7 @@ fn setup_head(
 				let header = batch.get_block_header(&head.last_block_h)?;
 
 				let res = txhashset::extending(header_pmmr, txhashset, &mut batch, |ext, batch| {
-					pipe::rewind_and_apply_fork(&header, ext, batch)?;
+					pipe::rewind_and_apply_fork(&header, ext, batch, &|_| Ok(()))?;
 
 					let extension = &mut ext.extension;
 
@@ -1892,7 +1668,7 @@ fn setup_head(
 					let prev_header = batch.get_block_header(&head.prev_block_h)?;
 
 					txhashset::extending(header_pmmr, txhashset, &mut batch, |ext, batch| {
-						pipe::rewind_and_apply_fork(&prev_header, ext, batch)
+						pipe::rewind_and_apply_fork(&prev_header, ext, batch, &|_| Ok(()))
 					})?;
 
 					// Now "undo" the latest block and forget it ever existed.
